@@ -24,6 +24,10 @@ class Connection(object):
     def url(self, url):
         self._url = url
 
+    @property
+    def version(self):
+        return self.make_api_request('GET', '/version')
+
     def make_api_request(self, http_method, url, data=None,
                          files=None, params=None):
         http_function = {
