@@ -1,8 +1,5 @@
 
 import StringIO
-from functools import wraps
-import math
-import time
 
 from pybamboo.connection import Connection
 from pybamboo.decorators import require_valid, retry
@@ -16,6 +13,7 @@ class Dataset(object):
     """
 
     _id = None
+    NA_VALUES = []
     NUM_RETRIES = 3.0
     AGGREGATIONS = [
         'max',
@@ -35,6 +33,7 @@ class Dataset(object):
     def __init__(self, dataset_id=None, url=None,
                  path=None, content=None, data_format='csv',
                  schema_path=None, schema_content=None,
+                 na_values=None,
                  connection=None):
         """
         Create a new pybamboo.Dataset from one of the following:
@@ -60,6 +59,16 @@ class Dataset(object):
                                     ' must be one of %s' %
                                     (data_format, self.DATA_FORMATS))
 
+        req_data = {}
+        if na_values is not None:
+            if not isinstance(na_values, (list, tuple, set)):
+                raise PyBambooException('N/A values must be a list.')
+            self.NA_VALUES = na_values
+            req_data.update({'na_values':
+                             safe_json_dumps(na_values,
+                                             PyBambooException('na_values '
+                                             'are not JSON-serializable'))})
+
         if connection is None:
             self._connection = Connection()
         else:
@@ -72,9 +81,9 @@ class Dataset(object):
 
         if url is not None:
             # TODO: check valid url?
-            data = {'url': url}
+            req_data.update({'url': url})
             self._id = self._connection.make_api_request(
-                'POST', '/datasets', data).get('id')
+                'POST', '/datasets', req_data).get('id')
             return
 
         # files might be overloaded by schema or path/content
@@ -93,39 +102,8 @@ class Dataset(object):
                          ('data.%s' % data_format, data)})
 
         self._id = self._connection.make_api_request('POST', '/datasets',
-                                                     files=files).get('id')
-
-    def _split_formula(self, formula):
-        return [token.strip() for token in formula.split('=', 1)]
-
-    def _is_formula_an_aggregation(self, formula):
-        for aggregation in self.AGGREGATIONS:
-            if formula.startswith(aggregation):
-                return True
-        return False
-
-    def _process_formula(self, formula, is_aggregation=False, as_dict=False):
-        try:
-            name, formula = self._split_formula(formula)
-            if not is_aggregation:
-                if self._is_formula_an_aggregation(formula):
-                    raise PyBambooException('"%s" is an aggregation. '
-                                            'Use Dataset.add_aggregation() '
-                                            'instead.' % formula)
-            else:
-                if not self._is_formula_an_aggregation(formula):
-                    raise PyBambooException('"%s" is a calculation. '
-                                            'Use Dataset.add_calculation() '
-                                            'instead.' % formula)
-            if as_dict:
-                return {'name': name, 'formula': formula}
-            return name, formula
-        except ValueError:
-            raise PyBambooException(
-                'Formulas must be of the form: "name = function".')
-        except AttributeError:
-            raise PyBambooException(
-                'Formulas must be expressed as strings.')
+                                                     files=files,
+                                                     data=req_data).get('id')
 
     def delete(self, num_retries=NUM_RETRIES):
         """
@@ -142,49 +120,87 @@ class Dataset(object):
             return success
         return _delete(self)
 
-    def add_calculation(self, formula, num_retries=NUM_RETRIES):
+    def add_calculation(self, name=None, formula=None, groups=None,
+                        num_retries=NUM_RETRIES):
         """
         Adds a calculation to this dataset in bamboo.
-        """
-        return self.add_calculations([formula], num_retries)
 
-    def add_calculations(self, formulae, num_retries=NUM_RETRIES):
-        """
-        Adds a list of calculations to this dataset in bamboo.
+        A calculation is defined by:
+        :param str name: the formula name by which you'll identify it.
+        :param str formula: the actual formula using bamboo syntax.
+            cf. http://bamboo.io/docs/index.html#formula-reference
+        :param list groups: optionnal, a list of fields to group on.
+
+        .. note ::
+            http://bamboo.io/docs/basic_commands.html#calculation-formulas
         """
         @require_valid
         @retry(num_retries)
-        def _add_calculations(self, formulae):
-            data = [self._process_formula(formula, as_dict=True)
-                    for formula in formulae]
-            json_data = safe_json_dumps(
-                data,
-                PyBambooException('formulae are not JSON-serializable'))
-            files = {'json_file': ('formulae.json', json_data)}
-            response = self._connection.make_api_request(
-                'POST', '/datasets/%s/calculations' % self._id, files=files)
-            return 'error' not in response.keys()
-        return _add_calculations(self, formulae)
+        def _add_calculation(self, formula, name, groups):
+            if (formula is None or name is None
+                or not isinstance(formula, basestring)
+                or not isinstance(name, basestring)):
+                raise PyBambooException('name & formula must be strings.')
 
-    def add_aggregation(self, formula, groups=None, num_retries=NUM_RETRIES):
-        """
-        Adds an aggregation to this dataset in bamboo.
-        """
-        @require_valid
-        @retry(num_retries)
-        def _add_aggregation(self, formula, groups):
-            name, formula = self._process_formula(formula,
-                                                  is_aggregation=True)
             data = {'name': name, 'formula': formula}
+
             if groups is not None:
                 if not isinstance(groups, list):
-                    raise PyBambooException(
-                        'groups must be a list of strings.')
+                    raise PyBambooException('group must be a list of strings.')
                 data['group'] = ','.join(groups)
+
             response = self._connection.make_api_request(
                 'POST', '/datasets/%s/calculations' % self._id, data=data)
             return 'error' not in response.keys()
-        return _add_aggregation(self, formula, groups)
+        return _add_calculation(self, formula, name, groups)
+
+    def add_calculations(self,
+                         path=None,
+                         content=None,
+                         json=None,
+                         num_retries=NUM_RETRIES):
+        """
+        Adds a JSON-formatted list of calculations to this dataset in bamboo.
+
+        Calculations are defined by a name, a formula and an optionnal grouping
+        This method takes a JSON list of dict, each containing:
+        * str name
+        * str formula
+        * list group (optionnal)
+
+        JSON content can be send in 3 ways:
+        :param str path: localation of a valid JSON file.
+        :param StringIO content: a file-like object containing valid JSON
+        :param dict json: a Python list of dict to be serialized.
+
+        .. note ::
+            http://bamboo.io/docs/advanced_commands.html
+                                  #creating-multiple-calculations-via-json
+        """
+        @require_valid
+        @retry(num_retries)
+        def _add_calculations(self, path, content, json):
+
+            files = {}
+
+            if path is None and content is None and json is None:
+                raise PyBambooException('JSON fomulae must be provided as '
+                                        'JSON, path or content.')
+
+            if json is not None:
+                json_data = safe_json_dumps(json,
+                                            PyBambooException('formulae are '
+                                                              'not JSON-serial'
+                                                              'izable'))
+                content = StringIO.StringIO(json_data)
+            data = content if content is not None else open(path)
+
+            files.update({'json_file': ('data.json', data)})
+
+            response = self._connection.make_api_request(
+                'POST', '/datasets/%s/calculations' % self._id, files=files)
+            return 'error' not in response.keys()
+        return _add_calculations(self, path, content, json)
 
     def remove_calculation(self, name, num_retries=NUM_RETRIES):
         """
@@ -198,12 +214,6 @@ class Dataset(object):
             return 'success' in response.keys()
         return _remove_calculation(self, name)
 
-    def remove_aggregation(self, name, num_retries=NUM_RETRIES):
-        """
-        Removes an aggregation from this dataset in bamboo.
-        """
-        return self.remove_calculation(name, num_retries)
-
     @require_valid
     def get_calculations(self):
         """
@@ -223,6 +233,10 @@ class Dataset(object):
         return dict([(group, Dataset(dataset_id, connection=self._connection))
                      for group, dataset_id in response.iteritems()])
 
+    @require_valid
+    def get_aggregations(self):
+        return self.get_aggregate_datasets()
+
     def get_summary(self, select='all', groups=None, query=None,
                     order_by=None, limit=0, callback=None,
                     num_retries=NUM_RETRIES):
@@ -291,8 +305,39 @@ class Dataset(object):
                 'GET', '/datasets/%s/info' % self._id, params=params)
         return _get_info(self, callback)
 
+    def set_info(self, attribution=None, description=None,
+                 label=None, license=None,
+                 num_retries=NUM_RETRIES):
+        """
+        Set metadata on the dataset
+        """
+        @require_valid
+        @retry(num_retries)
+        def _set_info(self, attribution, description, label, license):
+            params = {}
+            if attribution is not None:
+                if not isinstance(attribution, basestring):
+                    raise PyBambooException('attribution must be a string.')
+                params['attribution'] = attribution
+            if description is not None:
+                if not isinstance(description, basestring):
+                    raise PyBambooException('description must be a string.')
+                params['description'] = description
+            if label is not None:
+                if not isinstance(label, basestring):
+                    raise PyBambooException('label must be a string.')
+                params['label'] = label
+            if license is not None:
+                if not isinstance(license, basestring):
+                    raise PyBambooException('license must be a string.')
+                params['license'] = license
+            return self._connection.make_api_request(
+                'PUT', '/datasets/%s/info' % self._id, data=params)
+        return _set_info(self, attribution, description, label, license)
+
     def get_data(self, select=None, query=None, order_by=None, limit=0,
                  distinct=None, format=None, callback=None, count=False,
+                 index=False,
                  num_retries=NUM_RETRIES):
         """
         Returns the rows in this dataset filtered by the given
@@ -300,7 +345,7 @@ class Dataset(object):
         """
         @require_valid
         def _get_data(self, select, query, order_by, limit, distinct,
-                      format, callback, count):
+                      format, callback, count, index):
             params = {}
             if select:
                 if not isinstance(select, list):
@@ -340,10 +385,102 @@ class Dataset(object):
                     PyBambooException('limit is not JSON-serializable.'))
             if count:
                 params['count'] = bool(count)
+            if index:
+                params['index'] = bool(index)
             return self._connection.make_api_request(
                 'GET', '/datasets/%s' % self._id, params=params)
         return _get_data(self, select, query, order_by, limit, distinct,
-                         format, callback, count)
+                         format, callback, count, index)
+
+    def resample(self, date_column=None, interval=None, how=None,
+                 query=None, format=None):
+        """
+        Returns the rows in this dataset resampled by a date column.
+
+        The parameters are
+
+            date_column: The date column to resample on.
+            interval: A code for the interval to use.
+                      Any pandas codes are accepted.
+                      e.g. 'D' for daily, 'W' for weekly, 'M' for monthly.
+                      Complete list: http://pytseries.sourceforge.net
+                                          /core.constants.html#date-frequencies
+            how: (Optional) How to calculate the grouped samples.
+                      e.g: sum, mean, std, max, min, median, first, last, ohlc
+                      or any function/numpy array function. default is 'mean'.
+            query: (Optional) A MongoDB query to restrict the dataset,
+                      to only data matching the query will be resampled.
+            format: (Optional) format of the resampled data (CSV/JSON).
+
+        """
+        @require_valid
+        def _resample(self, date_column, interval, how, query, format):
+            params = {}
+            if not date_column or not isinstance(date_column, basestring):
+                raise PyBambooException('date_column must be a string.')
+
+            params['date_column'] = date_column
+
+            if not interval or not isinstance(interval, basestring):
+                raise PyBambooException('interval must be a string '
+                                        'representing a frequency. cf. '
+                                        'http://pytseries.sourceforge.net/'
+                                        'core.constants.html#date-frequencies')
+            params['interval'] = interval
+
+            if how:
+                if not isinstance(how, basestring):
+                    raise PyBambooException('how must be a string.')
+                params['how'] = how
+
+            if format:
+                if not isinstance(format, basestring):
+                    raise PyBambooException('format must be a string.')
+                params['format'] = format
+
+            if query:
+                if not isinstance(query, dict):
+                    raise PyBambooException('query must be a dict.')
+                params['query'] = safe_json_dumps(
+                    query,
+                    PyBambooException('query is not JSON-serializable.'))
+
+            return self._connection.make_api_request(
+                'GET', '/datasets/%s/resample' % self._id, params=params)
+        return _resample(self, date_column, interval, how, query, format)
+
+    def rolling(self, win_type=None, window=None, format=None):
+        """
+            To compute moving or rolling statistics / moments
+
+            you can use the rolling request.
+            Any options that can be passed to the pandas
+            rolling_window function can be passed as parameters to bamboo.
+            http://pandas.pydata.org/pandas-docs/dev
+                /computation.html#moving-rolling-statistics-moments
+            Window types are passed as the win_type parameter.
+        """
+
+        @require_valid
+        def _rolling(self, win_type, window, format):
+            params = {}
+            if win_type:
+                if not isinstance(win_type, basestring):
+                    raise PyBambooException('win_type must be a string.')
+                params['win_type'] = win_type
+
+            if not window or not isinstance(window, int):
+                raise PyBambooException('window must be an int')
+            params['window'] = window
+
+            if format:
+                if not isinstance(format, basestring):
+                    raise PyBambooException('format must be a string.')
+                params['format'] = format
+
+            return self._connection.make_api_request(
+                'GET', '/datasets/%s/rolling' % self._id, params=params)
+        return _rolling(self, win_type, window, format)
 
     @require_valid
     def update_data(self, rows):
@@ -441,6 +578,34 @@ class Dataset(object):
         else:
             return sum((int(relval) for relval in value.values()))
 
+    @require_valid
+    def row(self, action=None, index=None, payload=None):
+        data = None
+        if not action in ('show', 'delete', 'edit'):
+            raise PyBambooException('Row action must be show|edit|delete.')
+        if not isinstance(index, int):
+            raise PyBambooException('index must be an int.')
+        if action == 'edit':
+            data = {"data": safe_json_dumps(payload, PyBambooException(
+                    'payload is not JSON-serializable'))}
+        http_action = {'show': 'GET',
+                       'delete': 'DELETE',
+                       'edit': 'PUT'}.get(action)
+        return self._connection.make_api_request(
+            http_action, '/datasets/%s/row/%d' % (self._id, index), data=data)
+
+    @require_valid
+    def delete_row(self, index):
+        return self.row(action='delete', index=index)
+
+    @require_valid
+    def get_row(self, index):
+        return self.row(action='show', index=index)
+
+    @require_valid
+    def update_row(self, index, data):
+        return self.row(action='edit', index=index, payload=data)
+
     @property
     def id(self):
         """
